/*
   This is an implementation of the IntBag abstract data type (ADT).
   This class provides all the details of HOW an IntBag does what
   it does. In particular, this class specifies how the actual data
   will be stored (as an array).
*/

/**
* An IntArrayBag is a collection of int numbers.
* The same number may appear multiple times in a bag.
*
* @note
*   (1) The capacity of one of these bags can change after it's created, but
*   the maximum capacity is limited by the amount of free memory on the
*   machine. The constructor, addItem, clone,
*   and union will result in an OutOfMemoryError
*   when free memory is exhausted.
*   <p>
*   (2) A bag's capacity cannot exceed the maximum integer 2,147,483,647
*   (Integer.MAX_VALUE). Any attempt to create a larger capacity
*   results in a failure due to an arithmetic overflow.
*   <p>
*   (3) Because of the slow linear algorithms of this
*   class, large bags will have poor performance.
*
* @see IntArraysBag
*/
public class IntArrayBag extends IntBag
{
   // Invariant of the IntArrayBag class:
   //   1. The number of elements in the bag is in the instance variable
   //      manyItems, which is no more than data.length.
   //   2. For an empty bag, we do not care what is stored in any of data;
   //      for a non-empty bag, the elements in the bag are stored in data[0]
   //      through data[manyItems-1], and we don't care whats in the
   //      rest of data.
   private int[ ] data;
   private int manyItems;

   /**
   * Initialize an empty bag with an initial capacity of 10.  Note that the
   * addItem method works efficiently (without needing more
   * memory) until this capacity is reached.
   * @param - none
   * @postcondition
   *   This bag is empty and has an initial capacity of 10.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for:
   *   new int[10].
   */
   public IntArrayBag( )
   {
      final int INITIAL_CAPACITY = 10;
      manyItems = 0;
      data = new int[INITIAL_CAPACITY];
   }


   /**
   * Initialize an empty bag with a specified initial capacity. Note that the
   * addItem method works efficiently (without needing more
   * memory) until this capacity is reached.
   * @param initialCapacity
   *   the initial capacity of this bag
   * @precondition
   *   initialCapacity is non-negative.
   * @postcondition
   *   This bag is empty and has the given initial capacity.
   * @exception IllegalArgumentException
   *   Indicates that initialCapacity is negative.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for: new int[initialCapacity].
   */
   public IntArrayBag(int initialCapacity)
   {
      if (initialCapacity < 0)
         throw new IllegalArgumentException
         ("The initialCapacity is negative: " + initialCapacity);
      data = new int[initialCapacity];
      manyItems = 0;
   }


   /**
   * Change the current capacity of this bag.
   * @param minimumCapacity
   *   the new capacity for this bag
   * @postcondition
   *   This bag's capacity has been changed to at least minimumCapacity.
   *   If the capacity was already at or greater than minimumCapacity,
   *   then the capacity is left unchanged.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for: new int[minimumCapacity].
   */
   public void ensureCapacity(int minimumCapacity)
   {
      int[ ] biggerArray;

      if (data.length < minimumCapacity)
      {
         biggerArray = new int[minimumCapacity];
         System.arraycopy(data, 0, biggerArray, 0, manyItems);
         data = biggerArray;
      }
   }


   /**
   * Accessor method to get the current capacity of this bag.
   * The add method works efficiently (without needing
   * more memory) until this capacity is reached.
   * @param - none
   * @return
   *   the current capacity of this bag
   */
   public int getCapacity( )
   {
      return data.length;
   }


   /**
   * Reduce the current capacity of this bag to its actual size (i.e., the
   * number of elements it contains).
   * @param - none
   * @postcondition
   *   This bag's capacity has been changed to its current size.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for altering the capacity.
   */
   public void trimToSize( )
   {
      int[ ] trimmedArray;

      if (data.length != manyItems)
      {
         trimmedArray = new int[manyItems];
         System.arraycopy(data, 0, trimmedArray, 0, manyItems);
         data = trimmedArray;
      }
   }


   /**
   * {@inheritDoc}
   */
   public void add(int... elements)
   {
      if (manyItems + elements.length > data.length)
      {  // Ensure twice as much space as we need.
         ensureCapacity((manyItems + elements.length)*2);
      }

      System.arraycopy(elements, 0, data, manyItems, elements.length);
      manyItems += elements.length;
   }


   /**
   * {@inheritDoc}
   */
   public void add(IntBag addend)
   {
      // If addend is null, then a NullPointerException is thrown.
      ensureCapacity(manyItems + addend.size());

      System.arraycopy(addend.toArray(), 0, data, manyItems, addend.size());
      manyItems += addend.size();

      // What if addend == this? (See page 142 of the textbook.)
   }


   /**
   * {@inheritDoc}
   */
   public int countOccurrences(int target)
   {
      int result;
      int index;

      result = 0;
      for (index = 0; index < manyItems; index++)
      {
         if (target == data[index])
         {
            result++;
         }
      }
      return result;
   }


   /**
   * {@inheritDoc}
   */
   public boolean remove(int target)
   {
      int index; // The location of target in the data array.

      // First, set index to the location of target in the data array,
      // which could be as small as 0 or as large as manyItems - 1.
      // If target is not in the array, then index will be set equal
      // to manyItems;
      for (index = 0; (index < manyItems) && (target != data[index]); index++)
         /* No work is needed in the body of this for-loop. */
         ;

      if (index == manyItems)
      {  // The target was not found, so nothing is removed.
         return false;
      }
      else
      {  // The target was found at data[index].
         manyItems--;
         // copy the last element onto data[index].
         data[index] = data[manyItems];
         return true;
      }
   }


   /**
   * {@inheritDoc}
   */
   public int size( )
   {
      return manyItems;
   }


   /**
   * {@inheritDoc}
   */
   public IntBag union(IntBag b2)
   {
      // If b2 is null, then a NullPointerException is thrown.
      IntArrayBag result = new IntArrayBag(this.manyItems + b2.size());

      System.arraycopy(this.data,    0, result.data, 0,            this.manyItems);
      System.arraycopy(b2.toArray(), 0, result.data, this.manyItems, b2.size());
      result.manyItems = this.manyItems + b2.size();
/*
      result.add( this.toArray() );
      result.add(   b2.toArray() );
*/
      return result;

      // What if b2 == this? (See page 142 of the textbook.)
   }


   /**
   * {@inheritDoc}
   */
   public int[] toArray()
   {
      int[] result = new int[manyItems];
      for (int i = 0; i < manyItems; i++)
      {
         result[i] = data[i];
      }
      return result;
   }


   /**
   * {@inheritDoc}
   */
   public String toString()
   {
      String result = "[";
      for (int i = 0; i < manyItems; i++)
      {
         if (i > 0)
         {
            result += " ";
         }
         result += data[i];
         if (i < manyItems-1)
         {
            result += ",";
         }
      }
      result += "]";
      return result;
   }
/*
   public String toString()
   {
      return this.toArray().toString();
   }
*/

}
