/*
   This is an implementation of the IntBag abstract data type (ADT).
   This class provides all the details of HOW an IntBag does what
   it does. In particular, this class specifies how the actual data
   will be stored (as two arrays).
*/

/**
* An IntArraysBag is a collection of int numbers.
* The same number may appear multiple times in a bag.
*
* @note
*   (1) The capacity of one of these bags can change after it's created, but
*   the maximum capacity is limited by the amount of free memory on the
*   machine. The constructor, addItem, clone,
*   and union will result in an OutOfMemoryError
*   when free memory is exhausted.
*   <p>
*   (2) A bag's capacity cannot exceed the maximum integer 2,147,483,647
*   (Integer.MAX_VALUE). Any attempt to create a larger capacity
*   results in a failure due to an arithmetic overflow.
*   <p>
*   (3) Because of the slow linear algorithms of this
*   class, large bags will have poor performance.
*
* @see IntArrayBag
*/
public class IntArraysBag extends IntBag
{
   // The two arrays data and dataCounts are "parallel arrays"
   // The data[] array contains the numbers that are in the bag.
   // The dataCounts[] array  contains the count of the occurrences
   // of each number in the bag.
   // Say data[i] is 5 and dataCounts[i] is 12.
   // That means the bag holds 12 occurrences of the number 5.
   // Note that if data[i] is 5, then data[j] cannot be 5 for any j != i.
   private int[ ] data;
   private int[ ] dataCounts;
   private int manyItems;      // the number of entries in the bag
   private int distinctItems;  // the number of entries in data[]


   /**
   * Initialize an empty bag with an initial capacity of 10 distinct numbers.
   *
   * @param - none
   *
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for:
   *   new int[10].
   */
   public IntArraysBag( )
   {
      final int INITIAL_CAPACITY = 10;
      manyItems = 0;
      distinctItems = 0;
      data       = new int[INITIAL_CAPACITY];
      dataCounts = new int[INITIAL_CAPACITY];
   }


   /**
   * Initialize an empty bag with a specified initial capacity for distinct numbers.
   *
   * @param initialCapacity
   *   the initial capacity of distinct numbers in this bag
   *
   * @exception IllegalArgumentException
   *   Indicates that initialCapacity is negative.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for: new int[initialCapacity].
   */
   public IntArraysBag(int initialCapacity)
   {
      if (initialCapacity < 0)
         throw new IllegalArgumentException
         ("The initialCapacity is negative: " + initialCapacity);
      data       = new int[initialCapacity];
      dataCounts = new int[initialCapacity];
      manyItems = 0;
      distinctItems = 0;
   }


   /**
   * Change the current capacity of distinct numbers in this bag.
   * @param minimumCapacity
   *   the new capacity of distinct numbers for this bag
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for: new int[minimumCapacity].
   */
   public void ensureCapacity(int minimumCapacity)
   {
      int[ ] biggerArray1;
      int[ ] biggerArray2;

      if (data.length < minimumCapacity)
      {
         biggerArray1 = new int[minimumCapacity];
         System.arraycopy(data, 0, biggerArray1, 0, distinctItems);
         data = biggerArray1;

         biggerArray2 = new int[minimumCapacity];
         System.arraycopy(dataCounts, 0, biggerArray2, 0, distinctItems);
         dataCounts = biggerArray2;
      }
   }


   /**
   * Accessor method to get the current capacity of distinct numbers for this bag.
   * @param - none
   * @return
   *   the current capacity of this bag
   */
   public int getCapacity( )
   {
      return data.length;
   }


   /**
   * {@inheritDoc}
   */
   public void add(int... elements)
   {
      for (int element : elements)
      {
         int index = 0;
         for (; index < distinctItems; index++)
         {
            if (element == data[index])
               break;
         }

         if (index == distinctItems)
         {  // element is not in the bag
            if (distinctItems == data.length)
            {  // Ensure twice as much space as we need.
               ensureCapacity( (distinctItems + 1)*2 );
            }

            data[distinctItems] = element;
            dataCounts[distinctItems] = 1;
            distinctItems += 1;
            manyItems += 1;
         }
         else
         {  // element is already in the bag
            dataCounts[index] += 1;
            manyItems += 1;
         }
      }
   }


   /**
   * {@inheritDoc}
   */
   public void add(IntBag addend)
   {
      // If addend is null, then a NullPointerException is thrown.
      this.add( addend.toArray() );

      // What if addend == this? (See page 142 of the textbook.)
   }


   /**
   * {@inheritDoc}
   */
   public int countOccurrences(int target)
   {
      for (int index = 0; index < distinctItems; index++)
      {
         if (target == data[index])
            return dataCounts[index];
      }
      return 0;
   }


   /**
   * {@inheritDoc}
   */
   public boolean remove(int target)
   {
      int index; // The location of target in the data array.

      // First, set index to the location of target in the data array,
      // which could be as small as 0 or as large as distinctItems - 1.
      // If target is not in the array, then index will be set equal
      // to distinctItems;
      for (index = 0; (index < distinctItems) && (target != data[index]); index++)
         /* No work is needed in the body of this for-loop. */
         ;

      if (index == distinctItems)
      {  // The target was not found, so nothing is removed.
         return false;
      }
      else
      {  // The target was found at data[index].
         manyItems--;
         dataCounts[index]--;
         if ( 0 == dataCounts[index] )
         {
            distinctItems--;
            data[index]       = data[distinctItems];
            dataCounts[index] = dataCounts[distinctItems];
         }
         return true;
      }
   }


   /**
   * {@inheritDoc}
   */
   public int size( )
   {
      return manyItems;
   }


   /**
   * {@inheritDoc}
   */
   public IntBag union(IntBag b2)
   {
      // If b2 is null, then a NullPointerException is thrown.
      IntArraysBag result = new IntArraysBag(this.size() + b2.size());

      result.add( this.toArray() );
      result.add(   b2.toArray() );
/*
      for (int i = 0; i < this.distinctItems; i++)
      {
         for (int j = 0; j < this.dataCounts[i]; j++)
         {
            result.add( this.data[i] );
         }
      }
      result.add( b2.toArray() );
*/
      return result;

      // What if b2 == this? (See page 142 of the textbook.)
   }


   /**
   * {@inheritDoc}
   */
   public int[] toArray()
   {
      int[] result = new int[manyItems];
      int k = 0;
      for (int i = 0; i < distinctItems; i++)
      {
         for (int j = 0; j < dataCounts[i]; j++)
         {
            result[k++] = data[i];
         }
      }
      return result;
   }


   /**
   * {@inheritDoc}
   */
   public String toString()
   {
      String result = "[";
      for (int i = 0; i < distinctItems; i++)
      {
         if (i > 0)
         {
            result += " ";
         }
         result += data[i] + " x " + dataCounts[i];
         if (i < distinctItems-1)
         {
            result += ",";
         }
      }
      result += "]";

      return result;
   }
/*
   public String toString()
   {
      return this.toArray().toString();
   }
*/
/*
   public String toString()
   {
      String result = "[";
      for (int i = 0; i < distinctItems; i++)
      {
         for (int j = 0; j < dataCounts[i]; j++)
         {
            result += " " + data[i];
            if ( (i < distinctItems) && (j < dataCounts[i] - 1))
            {
               result += ",";
            }
         }
      }
      result += "]";

      return result;
   }
*/
}
