/*
   This is a better way to define the IntBag abstract data type (ADT).
   First define an abstract base class that specifies the interface
   to the ADT. Then define (possibly several) implementations of the
   interface.

   This class specifies WHAT an IntBag should be able to do but does
   not contain any implementation details (that is, there is no HOW in
   this file). In particular this abstract class does not specify how
   the actual data will be stored (that detail is left to each
   implementation).
*/

/**
* An IntBag is a collection of int numbers.
* The same number may appear multiple times in a bag.
*/
public abstract class IntBag
{
   /**
   * Add new elements to this bag.
   * @param elements
   *   (a variable-arity argument)
   *   one or more new elements that are being inserted
   * @postcondition
   *   A new copy of the element has been added to this bag.
   */
   public abstract void add(int... elements);


   /**
   * Add the contents of another bag to this bag.
   * @param addend
   *   a bag whose contents will be added to this bag
   * @precondition
   *   The parameter, addend, is not null.
   * @postcondition
   *   The elements from addend have been added to this bag.
   * @exception NullPointerException
   *   Indicates that addend is null.
   */
   public abstract void add(IntBag addend);


   /**
   * Accessor method to count the number of occurrences of a particular element
   * in this bag.
   * @param target
   *   the element that needs to be counted
   * @return
   *   the number of times that target occurs in this bag
   */
   public abstract int countOccurrences(int target);


   /**
   * Remove one copy of a specified element from this bag.
   * @param target
   *   the element to remove from the bag
   * @return
   *   true if the target element was in this bag, false otherwise
   * @postcondition
   *   If target was found in the bag, then one copy of
   *   target has been removed and the method returns true.
   *   Otherwise the bag remains unchanged and the method returns false.
   */
   public abstract boolean remove(int target);


   /**
   * Determine the number of elements in this bag.
   * @param - none
   * @return
   *   the number of elements in this bag
   */
   public abstract int size( );


   /**
   * Create a new bag that contains all the elements from this bag and the other bag.
   * @param b2
   *   the second bag
   * @precondition
   *   b2 is not null.
   * @return
   *   the union of this bag and the bag b2
   * @exception NullPointerException
   *   Indicates that the argument b2 is null.
   */
   public abstract IntBag union(IntBag b2);
   // Java doesn't allow abstract static methods.
   // That is why this method should not be static.
   // If we do make it static, then we must define
   // an empty "stub" method and just hope that each
   // subclass overrides it.
   /*
   public static IntBag union(IntBag b1, IntBag b2)
   {
      // Java doesn't allow abstract static methods.
      // So we define a stub method and just hope
      // that each subclass overrides it.
      return null;
   }
   */


   /**
   * Create an array representation of the contents of this bag.
   * @return
   *   an integer array containing every element of this bag
   */
   public abstract int[] toArray();


   /**
   * Create a String representation of the contents of this bag.
   * @return
   *   a String representation of the contents of this bag
   */
   public abstract String toString();

}
